/**
 * ***************************************************************************** Institute for
 * System Programming of the Russian Academy of Sciences Copyright (c) 2009-2013 All rights
 * reserved. This program and the accompanying materials are made available under the terms of the
 * Eclipse Public License v1.0 which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * ****************************************************************************
 */
package com.unitesk.requality_git.comm;

import org.eclipse.core.resources.IFile;
import org.eclipse.core.resources.IProject;
import org.eclipse.core.resources.IResource;
import org.eclipse.core.resources.team.FileModificationValidationContext;
import org.eclipse.core.resources.team.FileModificationValidator;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.IStatus;
import org.eclipse.core.runtime.NullProgressMonitor;
import org.eclipse.egit.core.GitProvider;
import org.eclipse.egit.core.op.ConnectProviderOperation;
import org.eclipse.egit.ui.internal.push.AddRemoteWizard;
import org.eclipse.egit.ui.internal.push.PushBranchWizard;
import org.eclipse.egit.ui.internal.push.PushWizard;
import org.eclipse.jface.wizard.WizardDialog;
import org.eclipse.jgit.api.*;
import org.eclipse.jgit.api.ResetCommand.ResetType;
import org.eclipse.jgit.api.errors.CheckoutConflictException;
import org.eclipse.jgit.api.errors.ConcurrentRefUpdateException;
import org.eclipse.jgit.api.errors.GitAPIException;
import org.eclipse.jgit.api.errors.InvalidTagNameException;
import org.eclipse.jgit.api.errors.NoHeadException;
import org.eclipse.jgit.lib.Constants;
import org.eclipse.jgit.lib.ObjectId;
import org.eclipse.jgit.lib.ObjectLoader;
import org.eclipse.jgit.lib.Ref;
import org.eclipse.jgit.lib.Repository;
import org.eclipse.jgit.revwalk.RevCommit;
import org.eclipse.jgit.revwalk.RevTree;
import org.eclipse.jgit.storage.file.FileRepositoryBuilder;
import org.eclipse.jgit.treewalk.TreeWalk;
import org.eclipse.jgit.treewalk.filter.OrTreeFilter;
import org.eclipse.jgit.treewalk.filter.PathFilter;
import org.eclipse.swt.SWT;
import org.eclipse.swt.widgets.Display;
import org.eclipse.team.core.RepositoryProvider;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.nio.file.Files;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.UUID;
import java.util.stream.Collectors;
import java.util.stream.Stream;
import com.unitesk.requality.core.AttributesVersioningRule;
import com.unitesk.requality.core.NodeMapper;
import com.unitesk.requality.core.TreeNode;
import com.unitesk.requality.core.attribute.AttributeDefinition;
import com.unitesk.requality.eclipse.cvs.CVSException;
import com.unitesk.requality.eclipse.storageformatversion.LegalicyNodeMapper;
import com.unitesk.requality.nodetypes.Document;

public class GitManipulator {

  IProject project = null;
  Repository repo = null;
  Git jGit = null;
  GitProvider teamProvider = null;
  Ref revHEAD = null;
  String currRev = null;
  String repoURL = null;

  public GitManipulator(IProject proj) throws Exception {
    if (proj != null) init(proj);
  }

  private File workingPlace;
  public static GitManipulator create(IProject proj) throws Exception{
    GitManipulator gm = new GitManipulator(null);
    gm.project = proj;
    gm.jGit = new InitCommand().setDirectory(proj.getLocation().toFile()).call();

    //RepositoryProvider.map(proj, GitProvider.ID);
    

    String projectAbsolutePath = proj.getLocationURI().getPath().toString() + File.separator;
    gm.workingPlace = new java.io.File(projectAbsolutePath + ".reqcvs");
    if (!gm.workingPlace.exists())
      if (!gm.workingPlace.mkdir()) {
        throw new CVSException("Cannot create folder " + gm.workingPlace.getAbsolutePath());
      }
    gm.repo = FileRepositoryBuilder.create(new File(projectAbsolutePath, ".git"));
    //gm.repo = gm.teamProvider.getData().getRepositoryMapping(proj).getRepository();
    //gm.jGit = new Git(gm.repo);
//    gm.repoURL =
//        gm.teamProvider
//            .getData()
//            .getRepositoryMapping(proj)
//            .getGitDirAbsolutePath()
//            .toFile()
//            .getParent()
//            .toString();
    
//    if (gm.repoURL == null || gm.repoURL.length() == 0)
//      throw new Exception("Empty repository URL string");
//    if (gm.repoURL.charAt(gm.repoURL.length() - 1) == '/')
//      gm.repoURL = gm.repoURL.substring(0, gm.repoURL.length() - 1);
    gm.addIgnore(".reqcvs");
    proj.refreshLocal(1, null);
    AddCommand ac = gm.jGit.add();
    for(File f:proj.getLocation().toFile().listFiles())
      if(!f.getName().startsWith(".")||!f.isDirectory())
        ac.addFilepattern(f.getName()+(f.isDirectory()?"/":""));
    ac.call();
    gm.jGit.commit().setMessage("initial commit").call();
    gm.teamProvider = (GitProvider) RepositoryProvider.getProvider(proj);
    if(gm.teamProvider == null){
      new ConnectProviderOperation(proj).execute(null);
      gm.teamProvider = (GitProvider) RepositoryProvider.getProvider(proj);
    }
//    org.eclipse.egit.ui.internal.push.PushWizard arw = new PushWizard(gm.repo);
//    WizardDialog wd = new WizardDialog(Display.getDefault().getActiveShell(), arw);
//    if(wd.open()==SWT.OK)
//      return gm;
    return null;
  }
  private void addIgnore(String string) {
    File f = workingPlace.getParentFile().toPath().resolve(".gitignore").toFile();
    if(!f.exists())
      try {
        f.createNewFile();
      } catch (IOException e) {
        // TODO Auto-generated catch block
        e.printStackTrace();
      }
    FileWriter fw = null;
    BufferedWriter bsw = null;
    try {
      Stream<String> fc = Files.lines(f.toPath());
      fw = new FileWriter(f);
      bsw = new BufferedWriter(fw);
      boolean was = false;
      for(String s:fc.collect(Collectors.toSet())){
        bsw.append(s);
        if(s.equals(string))
          was = true;
      }
      if(!was)
        bsw.write(string);
    } catch (IOException e) {
      e.printStackTrace();
    }
    finally{
      if(bsw!=null)
        try {
          bsw.close();
        } catch (IOException e1) {
          // TODO Auto-generated catch block
          e1.printStackTrace();
        }
      if(fw != null)
        try {
          fw.close();
        } catch (IOException e) {
          // TODO Auto-generated catch block
          e.printStackTrace();
        }
    }
  }
  public void init(IProject proj) throws Exception {
    if (proj == null) throw new Exception("NULL project");
    project = proj;
    RepositoryProvider tProv = RepositoryProvider.getProvider(project);

    if (!(tProv instanceof GitProvider))
      throw new Exception("Team provider seems to be NOT EGit, but should BE EGit");
    
    teamProvider = (GitProvider) tProv;

    String projectAbsolutePath = project.getLocationURI().getPath().toString() + File.separator;
    workingPlace = new java.io.File(projectAbsolutePath + ".reqcvs");

    if (!workingPlace.exists())
      if (!workingPlace.mkdir()) {
        throw new CVSException("Cannot create folder " + workingPlace.getAbsolutePath());
      }

    repo = teamProvider.getData().getRepositoryMapping(project).getRepository();

    //TODO path
    repoURL =
        teamProvider
            .getData()
            .getRepositoryMapping(project)
            .getGitDirAbsolutePath()
            .toFile()
            .getParent()
            .toString();
    jGit = new Git(repo);

    if (repoURL == null || repoURL.length() == 0)
      throw new Exception("Empty repository URL string");
    if (repoURL.charAt(repoURL.length() - 1) == '/')
      repoURL = repoURL.substring(0, repoURL.length() - 1);
    addIgnore(".reqcvs");
    proj.refreshLocal(1, null);
  }
  public String getPath(IResource r){
    String str = r.getFullPath().toString();
    int i = 0;
    while (str.charAt(i) == '/' || str.charAt(i) == '\\') i++;
    str = str.substring(i);
    str = str.replaceFirst(project.getName(), "");
    i = 0;
    while (i< str.length()&& (str.charAt(i) == '/' || str.charAt(i) == '\\')) i++;
    str = str.substring(i);
    return str;
  }
  public List<VersionInfo> listRevisions(IResource r) throws Exception {
    List<VersionInfo> commitsList = new ArrayList<VersionInfo>();
    Set<String> wasv = new HashSet<>();
    String path = getPath(r), path2 = r.getProject().getName()+"/"+path;
    org.eclipse.core.resources.team.FileModificationValidator fmv = new FileModificationValidator(){
    
      @Override
      public IStatus validateEdit(IFile[] files, FileModificationValidationContext context) {
        // TODO Auto-generated method stub
        return null;
      }
    };
    LogCommand logCommand = null;
    if(path.isEmpty())
      logCommand = jGit.log()
      .add(jGit.getRepository().resolve(Constants.HEAD)).all();
    else
      logCommand = jGit.log()
        .add(jGit.getRepository().resolve(Constants.HEAD))
        .addPath(path).addPath(path2);
    //int gv = 0;
    for (RevCommit revCommit : logCommand.call()) {
      //gv++;
      RevTree tree = revCommit.getTree();
      // now try to find a specific file
      try (TreeWalk treeWalk = new TreeWalk(jGit.getRepository())) {
        treeWalk.addTree(tree);
        treeWalk.setRecursive(true);
        if(!path.isEmpty())
          treeWalk.setFilter(OrTreeFilter.create(PathFilter.create(path),PathFilter.create(path2)));
        if(path.isEmpty()){
          if(treeWalk.next())
            commitsList.add(new VersionInfo(r, null, revCommit.getId().getName(), revCommit.getShortMessage()));
        }else
        while(treeWalk.next()){
//          if (!treeWalk.next()) {
//          //  continue;
//          }
          ObjectId objectId = treeWalk.getObjectId(0);
          ObjectLoader loader = jGit.getRepository().open(objectId);
  
            // and then one can the loader to read the file 
          ByteArrayOutputStream baos = new ByteArrayOutputStream();
          loader.copyTo(baos);
          baos.close();
          String file = baos.toString("UTF-8");
          try{
            UUID uid = NodeMapper.getUUId(file);
            Map<String,AttributeDefinition> map = null;
            try{
              map = NodeMapper.getAttributes(uid, file);
            }catch(Exception e){
              try{
                map = LegalicyNodeMapper.getAttributes(uid, file);
              }catch(Exception e2){
                e.printStackTrace();
              }
            }
            String vers = "0*";
            String type = "";//not legal by default
            if(map!=null && map.containsKey(AttributesVersioningRule.ATTR_NODE_VERSION))
              vers = map.get(AttributesVersioningRule.ATTR_NODE_VERSION).getRawValue().getValue().toString();
            if(map!=null && map.containsKey(TreeNode.ATTR_TYPE))
              type = map.get(TreeNode.ATTR_TYPE).getRawValue().getValue().toString();
            boolean illeagalType = !AttributesVersioningRule.isLegal(type);
            if(!wasv.contains(vers)||illeagalType){
              if(illeagalType)
                vers = AttributesVersioningRule.VALUE_ILLEGAL;
              commitsList.add(new VersionInfo(r,vers, revCommit.getId().getName(), revCommit.getShortMessage()));
                  //vers+": "+revCommit.getShortMessage()+"["+revCommit.getId().getName()+"]");
              wasv.add(vers);
            }
          }catch (Exception e) {
            //e.printStackTrace();
            // TODO: handle exception
          }
        }
      }

    }
    return commitsList;
  }
  public void setCurrentRevision(String rev) throws Exception {
    currRev = rev;
  }

  public void commit(IFile file, String commitMessage, IProgressMonitor monitor) throws Exception {
    if (monitor == null) monitor = new NullProgressMonitor();

    monitor.beginTask("Committing changes", 1);
    // добавить файл в index для наката
    String str = file.getFullPath().toString();

    int i = 0;
    while (str.charAt(i) == '/' || str.charAt(i) == '\\') i++;
    str = str.substring(i);
    str = str.replaceFirst(project.getName(), "");
    i = 0;
    while (str.charAt(i) == '/' || str.charAt(i) == '\\') i++;
    str = str.substring(i);

    jGit.add().addFilepattern(str).call();

    if (monitor.isCanceled()) {
      monitor.done();
      throw new InterruptedException();
    }
    // накатить файл
    jGit.commit().setMessage(commitMessage).call();
    monitor.worked(1);
    monitor.done();
    // throw new Exception("Not implemented yet");

  }

  public void excludeDotReqCSVDir() throws Exception {
    if (this.jGit == null || this.project == null) {
      throw new Exception("The CVCL SVC client has not been properly / seccessfully initialized");
    }

    File gitIgnoreFile =
        new File(project.getLocationURI().getPath().toString() + File.separator + ".gitignore");

    if (!gitIgnoreFile.exists()) {
      gitIgnoreFile.createNewFile();
    }

    BufferedReader reader = new BufferedReader(new FileReader(gitIgnoreFile));
    String str = reader.readLine();
    while (str != null) {
      if (str.trim().equalsIgnoreCase(".cvcl/")) {
        reader.close();
        return;
      }
      str = reader.readLine();
    }
    reader.close();

    int c = '\n';

    if (gitIgnoreFile.length() > 0) {
      reader = new BufferedReader(new FileReader(gitIgnoreFile));
      reader.skip(gitIgnoreFile.length() - 1);
      c = reader.read();
      reader.close();
    }

    BufferedWriter writer = new BufferedWriter(new FileWriter(gitIgnoreFile, true));
    if (c != '\n' && c != '\r') writer.newLine();
    writer.write(".cvcl/");
    writer.newLine();
    writer.close();
  }

  public void reset() throws CheckoutConflictException, GitAPIException, IOException {
    CleanCommand cc = jGit.clean();
    cc.setCleanDirectories(true);
    cc.setForce(true);
    cc.setIgnore(true);
    cc.call();
    ResetCommand reset = jGit.reset();
    reset.setMode(ResetType.HARD);
    reset.call();
  }

  public void tag(String tag)
      throws ConcurrentRefUpdateException, InvalidTagNameException, NoHeadException,
          GitAPIException {
    jGit.tag().setName(tag).call();
  }

  public void push(String message) throws CheckoutConflictException, GitAPIException {
    jGit.add().addFilepattern(".").call();
    jGit.add().setUpdate(true).addFilepattern(".").call();
    CommitCommand cc = jGit.commit();
    cc.setMessage(message);
    cc.call();
    jGit.push().call();
    //jGit.push().call();
  }

  public void pull() throws CheckoutConflictException, GitAPIException {
    PullCommand pc = jGit.pull();
    pc.call();
    ResetCommand reset = jGit.reset();
    reset.setMode(ResetType.HARD);
    reset.call();
  }

  public void close() {
    jGit.close();
  }
  public Repository getRepo() {
    return jGit.getRepository();
  }
}
