/*
 * Copyright 2015 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.unitesk.requality_git.properties;

import org.eclipse.compare.CompareConfiguration;
import org.eclipse.compare.CompareUI;
import org.eclipse.core.resources.IProject;
import org.eclipse.egit.ui.internal.CompareUtils;
import org.eclipse.egit.ui.internal.merge.GitCompareEditorInput;
import org.eclipse.jface.viewers.AbstractTableViewer;
import org.eclipse.swt.SWT;
import org.eclipse.swt.events.MouseEvent;
import org.eclipse.swt.events.MouseListener;
import org.eclipse.swt.layout.GridData;
import org.eclipse.swt.layout.GridLayout;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Label;
import org.eclipse.swt.widgets.Table;
import org.eclipse.swt.widgets.TableColumn;
import org.eclipse.swt.widgets.TableItem;
import org.eclipse.ui.PlatformUI;
import org.eclipse.ui.views.properties.tabbed.TabbedPropertySheetPage;
import org.eclipse.ui.views.properties.tabbed.TabbedPropertySheetWidgetFactory;
import com.unitesk.requality.core.AttributesVersioningRule;
import com.unitesk.requality.core.ChangesSet;
import com.unitesk.requality.core.TreeNode;
import com.unitesk.requality.core.attribute.Attribute;
import com.unitesk.requality.eclipse.core.EclipseTracker;
import com.unitesk.requality.eclipse.views.properties.TreeNodePropertiesTab;
import com.unitesk.requality_git.comm.GitManipulator;
import com.unitesk.requality_git.comm.VersionInfo;

public class NodeVersions extends TreeNodePropertiesTab {
  private Table versions;
  private TreeNode req;
  private Label vers;
  private GitManipulator gm;
  private AbstractTableViewer dv;
  private java.util.List<VersionInfo> versI;
  public NodeVersions() {}

  @Override
  public void createControls(Composite parent, TabbedPropertySheetPage aTabbedPropertySheetPage) {
    super.createControls(parent, aTabbedPropertySheetPage);
    TabbedPropertySheetWidgetFactory wf = aTabbedPropertySheetPage.getWidgetFactory();

    parent.setLayout(new GridLayout());
    parent.setLayoutData(new GridData(GridData.FILL_BOTH));

    wf.createLabel(parent, "Current version:");
    String vers = "*";
    if(req!=null)
      vers = getNodeVersion(req);
    this.vers = wf.createLabel(parent, vers);
    this.vers.setLayoutData(new GridData(GridData.FILL_HORIZONTAL));
    versions = new Table( parent, SWT.READ_ONLY | SWT.BORDER | SWT.V_SCROLL);
//    versions = new List(parent, SWT.READ_ONLY | SWT.BORDER | SWT.V_SCROLL);
    versions.setLayoutData(new GridData(GridData.FILL_BOTH));
    versions.addMouseListener(new MouseListener(){
    
      @Override
      public void mouseUp(MouseEvent e) {
        // TODO Auto-generated method stub
        
      }
    
      @Override
      public void mouseDown(MouseEvent e) {
        // TODO Auto-generated method stub
        
      }
    
      @Override
      public void mouseDoubleClick(MouseEvent e) {
        int ind = versions.getSelectionIndex();
        if(ind == 0){
          if(!req.isRoot())
            CompareUtils.compareWorkspaceWithRef(gm.getRepo(), versI.get(ind).getResource(), versI.get(ind).getCommit(), 
              PlatformUI.getWorkbench().getActiveWorkbenchWindow().getActivePage());
        }
        else
        if(ind > 0)
          openCompare(versI.get(ind));
      }
    });
    String[] titles = { "Version", "Commit message",
        "Revision" };
    for (int i = 0; i < titles.length; i++) {
      TableColumn column = new TableColumn(versions, SWT.NONE);
      column.setText(titles[i]);
    }
    //dv = new org.eclipse.egit.ui.Activator.;(parent, PlatformUI.getWorkbench().getActiveWorkbenchWindow().getActivePage().getActivePart().getSite());
    
  }

  protected void openCompare(VersionInfo versionInfo) {
    CompareConfiguration c = new CompareConfiguration();
    if(req.isRoot())
      CompareUI.openCompareEditor(
          new GitCompareEditorInput(
              "HEAD",
              versionInfo.getCommit(), gm.getRepo()));
    else
      CompareUtils.compareWorkspaceWithRef(gm.getRepo(), versionInfo.getResource(), versionInfo.getCommit(), 
          PlatformUI.getWorkbench().getActiveWorkbenchWindow().getActivePage());
  }

  private String getNodeVersion(TreeNode req2) {
    return req2.getNodeVersion();
  }

  @Override
  protected boolean affectedByAttribute(String a) {
    return AttributesVersioningRule.ATTR_NODE_VERSION.equals(a);
  };

  @Override
  protected void setTreeNode(TreeNode node) {
    if(gm!=null)
      gm.close();
    req = node;

    if (req == null || versions == null || versions.isDisposed()) {
      return;
    }
    if(node.isRoot())
      vers.setText("");
    else
      vers.setText(getNodeVersion(req));
    IProject proj = EclipseTracker.getResourceByNode(req).getProject();
    try {
      gm = new GitManipulator(proj);
    }
    catch(Exception e){
      //e.printStackTrace();
    }
    versions.clearAll();
    //versions.setItems(new String[0]);
    if(gm == null)
      return;
    try {
      versI = gm.listRevisions(req.isRoot()?EclipseTracker.getResourceByNode(node).getProject():EclipseTracker.getResourceByNode(node));
//      versI.sort(new Comparator<VersionInfo>(){
//        @Override
//        public int compare(VersionInfo o1, VersionInfo o2) {
//          int i1 = 0, i2 = 0;
//          if(o1.getVersion().endsWith("*"))
//            i1 = Integer.parseInt(o1.getVersion().substring(0,o1.getVersion().length()-1));
//          else
//            i1 = Integer.parseInt(o1.getVersion());
//
//          if(o2.getVersion().endsWith("*"))
//            i2 = Integer.parseInt(o2.getVersion().substring(0,o2.getVersion().length()-1));
//          else
//            i2 = Integer.parseInt(o2.getVersion());
//          return Integer.compare(i2, i1);
//        }
//      });
      versions.clearAll();
      versions.setItemCount(0);
      for (VersionInfo vi:versI) {
        TableItem item = new TableItem(versions, SWT.NONE);
        item.setText(0, vi.getVersion()!=null ? vi.getVersion():"");
        item.setText(1, vi.getMessage());
        item.setText(2, vi.getCommit());
      }
      for (TableColumn s: versions.getColumns()) {
        s.pack();
      }
      //versions.setSize(versions.computeSize(SWT.DEFAULT, 200));
      
    } catch (Exception e) {
      e.printStackTrace();
    }
  }
  private static String[] toStringArray(java.util.List<VersionInfo> versI2) {
    String[]infos = new String[versI2.size()];
    int i = 0;
    for(VersionInfo vi:versI2)
      infos[i++] = vi.toString();
    return infos;
  }

  @Override
  public boolean requireUpdate(TreeNode source, ChangesSet<Attribute> what) {
    return what.getAll().contains(AttributesVersioningRule.ATTR_NODE_VERSION);
  };
}
